import unittest
from datetime import datetime

from discord_webhook import DiscordEmbed
from discord_webhook.webhook_exceptions import ColorNotInRangeException


class TestWebhookEmbed(unittest.TestCase):
    def setUp(self):
        self.embed = DiscordEmbed()

    def test_set_embed_title(self):
        embed_title = "foobar"
        self.embed.set_title(embed_title)
        self.assertEqual(self.embed.title, embed_title)

    def test_set_embed_description(self):
        embed_description = "just a short description"
        self.embed.set_description(embed_description)
        self.assertEqual(self.embed.description, embed_description)

    def test_set_embed_url(self):
        embed_url = "testurl"
        self.embed.set_url(embed_url)
        self.assertEqual(self.embed.url, embed_url)

    def test_set_embed_timestamp(self):
        embed_timestamp = 1658504418.5660195
        self.embed.set_timestamp(embed_timestamp)
        self.assertEqual(self.embed.timestamp,
                         str(datetime.utcfromtimestamp(embed_timestamp)))

    def test_set_embed_color(self):
        embed_color_as_str = "03b2f8"
        embed_color_as_int = int(embed_color_as_str, 16)
        self.embed.set_color(embed_color_as_str)
        self.assertEqual(self.embed.color, embed_color_as_int)
        self.embed.set_color(embed_color_as_int)
        self.assertEqual(self.embed.color, embed_color_as_int)

    def test_embed_color_out_of_range(self):
        embed_color = 9999999999
        with self.assertRaises(ColorNotInRangeException):
            self.embed.set_color(embed_color)

    def test_embed_set_footer(self):
        footer_text = "footer text"
        footer_icon_url = "footer icon url"
        footer_proxy_icon_url = "proxied footer icon url"
        self.embed.set_footer(text=footer_text, icon_url=footer_icon_url,
                              proxy_icon_url=footer_proxy_icon_url)
        self.assertEqual(self.embed.footer, {
            "text": footer_text,
            "icon_url": footer_icon_url,
            "proxy_icon_url": footer_proxy_icon_url
        })

    def test_embed_set_image(self):
        image_url = "image url"
        image_proxy_url = "image proxy url"
        image_height = 500
        image_width = 500
        self.embed.set_image(url=image_url, proxy_url=image_proxy_url,
                             height=image_height, width=image_width)
        self.assertEqual(self.embed.image, {
            "url": image_url,
            "proxy_url": image_proxy_url,
            "height": image_height,
            "width": image_width,
        })

    def test_embed_set_thumbnail(self):
        thumbnail_url = "thumbnail url"
        thumbnail_proxy_url = "thumbnail proxy url"
        thumbnail_height = 500
        thumbnail_width = 500
        self.embed.set_thumbnail(url=thumbnail_url,
                                 proxy_url=thumbnail_proxy_url,
                                 height=thumbnail_height, width=thumbnail_width)
        self.assertEqual(self.embed.thumbnail, {
            "url": thumbnail_url,
            "proxy_url": thumbnail_proxy_url,
            "height": thumbnail_height,
            "width": thumbnail_width,
        })

    def test_embed_set_video(self):
        video_url = "video url"
        video_height = 500
        video_width = 500
        self.embed.set_video(url=video_url, height=video_height,
                             width=video_width)
        self.assertEqual(self.embed.video, {
            "url": video_url,
            "height": video_height,
            "width": video_width,
        })

    def test_embed_set_provider(self):
        provider_name = "provider name"
        provider_url = "provider url"
        self.embed.set_provider(name=provider_name, url=provider_url)
        self.assertEqual(self.embed.provider, {
            "name": provider_name,
            "url": provider_url,
        })

    def test_embed_set_author(self):
        author_name = "author name"
        author_url = "author url"
        author_icon_url = "author icon url"
        author_proxy_icon_url = "author proxy icon url"
        self.embed.set_author(name=author_name, url=author_url,
                              icon_url=author_icon_url,
                              proxy_icon_url=author_proxy_icon_url)
        self.assertEqual(self.embed.author, {
            "name": author_name,
            "url": author_url,
            "icon_url": author_icon_url,
            "proxy_icon_url": author_proxy_icon_url,
        })

    def test_embed_set_field(self):
        field_counter = 0
        field_name = "field name"
        field_value = "field value"
        field_inline = False
        self.assertEqual(len(self.embed.fields), field_counter)
        self.embed.add_embed_field(name=field_name, value=field_value,
                                   inline=field_inline)
        self.assertEqual(len(self.embed.fields), field_counter + 1)
        self.assertEqual(self.embed.fields[0], {
                "name": field_name,
                "value": field_value,
                "inline": field_inline,
            })



import datetime
import json
import logging
import time
from functools import partial
from typing import Any, Dict, List, Optional, Tuple, Union, cast

import requests
from .webhook_exceptions import ColorNotInRangeException

logger = logging.getLogger(__name__)


class DiscordEmbed:
    """
    Discord Embed
    """

    title: Optional[str]
    description: Optional[str]
    url: Optional[str]
    timestamp: Optional[str]
    color: Optional[int]
    footer: Optional[Dict[str, Optional[str]]]
    image: Optional[Dict[str, Optional[Union[str, int]]]]
    thumbnail: Optional[Union[str, Dict[str, Optional[Union[str, int]]]]]
    video: Optional[Union[str, Dict[str, Optional[Union[str, int]]]]]
    provider: Optional[Dict[str, Any]]
    author: Optional[Dict[str, Optional[str]]]
    fields: List[Dict[str, Optional[Any]]]

    def __init__(
            self,
            title: Optional[str] = None,
            description: Optional[str] = None,
            **kwargs: Any,
    ) -> None:
        """
        Init Discord Embed
        -----------
        :keyword ``title:`` title of embed\n
        :keyword ``description:`` description body of embed\n
        :keyword ``url:`` add an url to make your embedded title a clickable
        link\n
        :keyword ``timestamp:`` timestamp of embed content\n
        :keyword ``color:`` color code of the embed as int\n
        :keyword ``footer:`` footer texts\n
        :keyword ``image:`` your image url here\n
        :keyword ``thumbnail:`` your thumbnail url here\n
        :keyword ``video:``  to apply video with embedded, your video source
        url here\n
        :keyword ``provider:`` provider information\n
        :keyword ``author:`` author information\n
        :keyword ``fields:`` fields information
        """
        self.title = title
        self.description = description
        self.url = cast(str, kwargs.get("url"))
        self.timestamp = cast(str, kwargs.get("timestamp"))
        self.footer = kwargs.get("footer")
        self.image = kwargs.get("image")
        self.thumbnail = kwargs.get("thumbnail")
        self.video = kwargs.get("video")
        self.provider = kwargs.get("provider")
        self.author = kwargs.get("author")
        self.fields = kwargs.get("fields", [])
        self.set_color(kwargs.get("color"))

    def set_title(self, title: str) -> None:
        """
        set title of embed
        :param title: title of embed
        """
        self.title = title

    def set_description(self, description: str) -> None:
        """
        set description of embed
        :param description: description of embed
        """
        self.description = description

    def set_url(self, url: str) -> None:
        """
        set url of embed
        :param url: url of embed
        """
        self.url = url

    def set_timestamp(self, timestamp: Optional[float] = None) -> None:
        """
        set timestamp of embed content
        :param timestamp: (optional) timestamp of embed content
        """
        if timestamp is None:
            timestamp = time.time()
        self.timestamp = str(datetime.datetime.utcfromtimestamp(timestamp))

    def set_color(self, color: Union[str, int]) -> None:
        """
        set color code of the embed as decimal(int) or hex(string)
        :param color: color code of the embed as decimal(int) or hex(string)
        """
        self.color = int(color, 16) if isinstance(color, str) else color
        if self.color is not None and self.color not in range(16777216):
            raise ColorNotInRangeException(color)

    def set_footer(self, **kwargs: str) -> None:
        """
        set footer information of embed
        :keyword text: footer text
        :keyword icon_url: url of footer icon (only supports http(s) and
        attachments)
        :keyword proxy_icon_url: a proxied url of footer icon
        """
        self.footer = {
            "text": kwargs.get("text"),
            "icon_url": kwargs.get("icon_url"),
            "proxy_icon_url": kwargs.get("proxy_icon_url"),
        }

    def set_image(self, url: str, **kwargs: Union[str, int]) -> None:
        """
        set image of embed
        :param url: source url of image (only supports http(s) and attachments)
        :keyword proxy_url: a proxied url of the image
        :keyword height: height of image
        :keyword width: width of image
        """
        self.image = {
            "url": url,
            "proxy_url": cast(Optional[str], kwargs.get("proxy_url")),
            "height": cast(Optional[int], kwargs.get("height")),
            "width": cast(Optional[int], kwargs.get("width")),
        }

    def set_thumbnail(self, url: str, **kwargs: Union[str, int]) -> None:
        """
        set thumbnail of embed
        :param url: source url of thumbnail (only supports http(s) and
        attachments)
        :keyword proxy_url: a proxied thumbnail of the image
        :keyword height: height of thumbnail
        :keyword width: width of thumbnail
        """
        self.thumbnail = {
            "url": url,
            "proxy_url": cast(Optional[str], kwargs.get("proxy_url")),
            "height": cast(Optional[int], kwargs.get("height")),
            "width": cast(Optional[str], kwargs.get("width")),
        }

    def set_video(self, **kwargs: Union[str, int]) -> None:
        """
        set video of embed
        :keyword url: source url of video
        :keyword height: height of video
        :keyword width: width of video
        """
        self.video = {
            "url": cast(Optional[str], kwargs.get("url")),
            "height": cast(Optional[int], kwargs.get("height")),
            "width": cast(Optional[int], kwargs.get("width")),
        }

    def set_provider(self, **kwargs: str) -> None:
        """
        set provider of embed
        :keyword name: name of provider
        :keyword url: url of provider
        """
        self.provider = {
            "name": kwargs.get("name"),
            "url": kwargs.get("url"),
        }

    def set_author(self, name: str, **kwargs: str) -> None:
        """
        set author of embed
        :param name: name of author
        :keyword url: url of author
        :keyword icon_url: url of author icon (only supports http(s) and
        attachments)
        :keyword proxy_icon_url: a proxied url of author icon
        """
        self.author = {
            "name": name,
            "url": kwargs.get("url"),
            "icon_url": kwargs.get("icon_url"),
            "proxy_icon_url": kwargs.get("proxy_icon_url"),
        }

    def add_embed_field(self, name: str, value: str,
                        inline: bool = True) -> None:
        """
        set field of embed
        :param name: name of the field
        :param value: value of the field
        :param inline: (optional) whether this field should display inline
        """
        self.fields.append(
            {
                "name": name,
                "value": value,
                "inline": inline,
            }
        )

    def del_embed_field(self, index: int) -> None:
        """
        remove field from `self.fields`
        :param index: index of field in `self.fields`
        """
        self.fields.pop(index)

    def get_embed_fields(self) -> List[Dict[str, Optional[Any]]]:
        """
        get all `self.fields` as list
        :return: `self.fields`
        """
        return self.fields


class DiscordWebhook:
    """
    Webhook for Discord
    """

    url: Optional[Union[str, List[str]]]
    content: Optional[Union[str, bytes]]
    username: Optional[str]
    avatar_url: Optional[str]
    tts: bool
    files: Dict[str, Tuple[Optional[str], Union[bytes, str]]]
    embeds: List[Dict[str, Any]]
    proxies: Optional[Dict[str, str]]
    allowed_mentions: List[str]
    timeout: Optional[float]
    rate_limit_retry: bool = False

    def __init__(
            self,
            url: Optional[Union[str, List[str]]] = None,
            *,
            content: Optional[str] = None,
            username: Optional[str] = None,
            avatar_url: Optional[str] = None,
            tts: bool = False,
            files: Optional[
                Dict[str, Tuple[Optional[str], Union[bytes, str]]]] = None,
            embeds: Optional[List[Dict[str, Any]]] = None,
            proxies: Optional[Dict[str, str]] = None,
            timeout: Optional[float] = None,
            rate_limit_retry: bool = False,
            allowed_mentions: Optional[List[str]] = None,
    ) -> None:
        """
        Init Webhook for Discord
        ---------
        :param ``url``: your discord webhook url (type: str, list)\n
        :keyword ``content:`` the message contents (type: str)\n
        :keyword ``username:`` override the default username of the webhook\n
        :keyword ``avatar_url:`` override the default avatar of the webhook\n
        :keyword ``tts:`` true if this is a TTS message\n
        :keyword ``file``: to apply file(s) with message
        (For example: file=f.read() (here, f = variable that contain
        attachement path as "rb" mode))\n
        :keyword ``filename:`` apply custom file name on attached file
        content(s)\n
        :keyword ``embeds:`` list of embedded rich content\n
        :keyword ``allowed_mentions:`` allowed mentions for the message\n
        :keyword ``proxies:`` dict of proxies\n
        :keyword ``timeout:`` (optional) amount of seconds to wait for a
        response from Discord
        """
        if embeds is None:
            embeds = []
        if files is None:
            files = {}
        if allowed_mentions is None:
            allowed_mentions = []
        self.url = url
        self.content = content
        self.username = username
        self.avatar_url = avatar_url
        self.tts = tts
        self.files = files
        self.embeds = embeds
        self.proxies = proxies
        self.allowed_mentions = allowed_mentions
        self.timeout = timeout
        self.rate_limit_retry = rate_limit_retry

    def add_file(self, file: bytes, filename: str) -> None:
        """
        adds a file to the webhook
        :param file: file content
        :param filename: filename
        :return:
        """
        self.files[f"_{filename}"] = (filename, file)

    def add_embed(self, embed: Union[DiscordEmbed, Dict[str, Any]]) -> None:
        """
        adds an embedded rich content
        :param embed: embed object or dict
        """
        self.embeds.append(
            embed.__dict__ if isinstance(embed, DiscordEmbed) else embed)

    def remove_embed(self, index: int) -> None:
        """
        removes embedded rich content from `self.embeds`
        :param index: index of embed in `self.embeds`
        """
        self.embeds.pop(index)

    def remove_file(self, filename: str) -> None:
        """
        removes file from `self.files` using specified `filename` if it exists
        :param filename: filename
        """
        filename = f"_{filename}"
        if filename in self.files:
            del self.files[filename]

    def get_embeds(self) -> List[Dict[str, Any]]:
        """
        gets all self.embeds as list
        :return: self.embeds
        """
        return self.embeds

    def set_proxies(self, proxies: Dict[str, str]) -> None:
        """
        sets proxies
        :param proxies: dict of proxies
        :type proxies: dict
        """
        self.proxies = proxies

    def set_content(self, content: str) -> None:
        """
        sets content
        :param content: content string
        :type content: string
        """
        self.content = content

    @property
    def json(self) -> Dict[str, Any]:
        """
        convert webhook data to json
        :return webhook data as json:
        """
        embeds = self.embeds
        self.embeds = []
        # convert DiscordEmbed to dict
        for embed in embeds:
            self.add_embed(embed)
        data = {
            key: value
            for key, value in self.__dict__.items()
            if value and key not in {"url", "files", "filename"}
        }
        embeds_empty = not any(data["embeds"]) if "embeds" in data else True
        if embeds_empty and "content" not in data and bool(self.files) is False:
            logger.error("webhook message is empty! set content or embed data")
        return data

    def remove_embeds(self) -> None:
        """
        Sets `self.embeds` to empty `list`.
        """
        self.embeds = []

    def remove_files(self) -> None:
        """
        Sets `self.files` to empty `dict`.
        """
        self.files = {}

    def api_post_request(self, url: str) -> requests.Response:
        if bool(self.files) is False:
            response = requests.post(
                url,
                json=self.json,
                proxies=self.proxies,
                params={"wait": True},
                timeout=self.timeout,
            )
        else:
            self.files["payload_json"] = (None, json.dumps(self.json))
            response = requests.post(
                url,
                files=self.files,
                proxies=self.proxies,
                timeout=self.timeout,
            )
        return response

    def execute(
            self,
            remove_embeds: bool = False,
            remove_files: bool = False,
    ) -> Union[List[requests.Response], requests.Response]:
        """
        executes the Webhook
        :param remove_embeds: if set to True, calls `self.remove_embeds()`
        to empty `self.embeds` after webhook is executed
        :param remove_files: if set to True, calls `self.remove_files()`
        to empty `self.files` after webhook is executed
        :return: Webhook response
        """
        webhook_urls = self.url
        if isinstance(self.url, str):
            webhook_urls = [self.url]
        urls_len = len(webhook_urls)
        responses = []
        for i, url in enumerate(webhook_urls):
            response = self.api_post_request(url)
            if response.status_code in [200, 204]:
                logger.debug(f"[{i + 1}/{urls_len}] Webhook executed")
            elif response.status_code == 429 and self.rate_limit_retry:
                while response.status_code == 429:
                    errors = json.loads(response.content.decode("utf-8"))
                    wh_sleep = (int(errors["retry_after"]) / 1000) + 0.15
                    time.sleep(wh_sleep)
                    logger.error(
                        f"Webhook rate limited: sleeping for {wh_sleep} " 
                        "seconds..."
                    )
                    response = self.api_post_request(url)
                    if response.status_code in [200, 204]:
                        logger.debug(f"[{i + 1}/{urls_len}] Webhook executed")
                        break
            else:
                logger.error(
                    f"[{i + 1}/{urls_len}] Webhook status code "
                    f"{response.status_code}: "
                    f"{response.content.decode('utf-8')}"
                )
            responses.append(response)
        if remove_embeds:
            self.remove_embeds()
        if remove_files:
            self.remove_files()
        return responses[0] if len(responses) == 1 else responses

    def edit(
            self,
            sent_webhook: Union[List[requests.Response], requests.Response],
    ) -> Union[List[requests.Response], requests.Response]:
        """
        edits the webhook passed as a response
        :param sent_webhook: webhook.execute() response
        :return: Another webhook response
        """
        if not isinstance(sent_webhook, list):
            sent_webhook = cast(List[requests.Response], [sent_webhook])
        responses: List[requests.Response] = []
        for i, webhook in enumerate(sent_webhook):
            assert isinstance(webhook.content, bytes)
            previous_sent_message_id = json.loads(
                webhook.content.decode("utf-8")
            )["id"]
            url = (
                    webhook.url.split("?")[0] + "/messages/" + str(
                previous_sent_message_id)
            )
            # removes any query params
            if bool(self.files) is False:
                request = partial(
                    requests.patch,
                    url,
                    json=self.json,
                    proxies=self.proxies,
                    params={"wait": True},
                    timeout=self.timeout,
                )
            else:
                self.files["payload_json"] = (None, json.dumps(self.json))
                request = partial(
                    requests.patch,
                    url,
                    files=self.files,
                    proxies=self.proxies,
                    timeout=self.timeout,
                )
            response = request()
            if response.status_code in [200, 204]:
                logger.debug(f"[{i + 1}/{len(sent_webhook)}] Webhook edited")
            elif response.status_code == 429 and self.rate_limit_retry:
                while response.status_code == 429:
                    errors = json.loads(response.content.decode("utf-8"))
                    wh_sleep = (int(errors["retry_after"]) / 1000) + 0.15
                    time.sleep(wh_sleep)
                    logger.error(
                        f"Webhook rate limited: sleeping for {wh_sleep} "
                        f"seconds..."
                    )
                    response = request()
                    if response.status_code in [200, 204]:
                        logger.debug(
                            f"[{i + 1}/{len(sent_webhook)}] Webhook edited")
                        break
            else:
                logger.error(
                    f"[{i + 1}/{len(sent_webhook)}] Webhook status code "
                    f"{response.status_code}: "
                    f"{response.content.decode('utf-8')}"
                )
            responses.append(response)
        return responses[0] if len(responses) == 1 else responses

    def delete(
            self, sent_webhook: Union[List["DiscordWebhook"], "DiscordWebhook"]
    ) -> Union[List[requests.Response], requests.Response]:
        """
        deletes the webhook passed as a response
        :param sent_webhook: webhook.execute() response
        :return: Response
        """
        if not isinstance(sent_webhook, list):
            sent_webhook = cast(List[DiscordWebhook], [sent_webhook])
        responses: List[requests.Response] = []
        for i, webhook in enumerate(sent_webhook):
            assert isinstance(webhook.content, bytes)
            url = webhook.url.split("?")[0]  # removes any query params
            previous_sent_message_id = json.loads(
                webhook.content.decode("utf-8")
            )["id"]
            response = requests.delete(
                url + "/messages/" + str(previous_sent_message_id),
                proxies=self.proxies,
                timeout=self.timeout,
            )
            if response.status_code in [200, 204]:
                logger.debug(f"[{i + 1}/{len(sent_webhook)}] Webhook deleted")
            else:
                logger.error(
                    f"[{i + 1}/{len(sent_webhook)}] Webhook status code "
                    f"{response.status_code}: {response.content.decode('utf-8')}"
                )
            responses.append(response)
        return responses[0] if len(responses) == 1 else responses



import asyncio
import json
import logging
from contextlib import asynccontextmanager

from . import DiscordWebhook

logger = logging.getLogger(__name__)

try:
    import httpx  # noqa
except ImportError:  # pragma: nocover
    # Async is an optional dependency so don't raise
    # an exception unless the AsyncDiscordWebhook is used.
    pass


class AsyncDiscordWebhook(DiscordWebhook):
    """
    Async version of DiscordWebhook.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        try:
            import httpx  # noqa
        except ImportError:  # pragma: nocover
            raise ImportError(
                "You're attempting to use the async version of discord-webhooks but didn't"
                " install it using `pip install discord-webhook[async]`."
            ) from None

    @property
    @asynccontextmanager
    async def http_client(self):
        """
        A property that returns an httpx.AsyncClient instance that is used for a 'with' statement.
        Example:
            async with self.http_client as client:
                client.post(url, data=data)
        It will automatically close the client when the context is exited.
        :return: httpx.AsyncClient
        """
        client = httpx.AsyncClient(proxies=self.proxies)
        yield client
        await client.aclose()

    async def api_post_request(self, url):
        async with self.http_client as client:  # type: httpx.AsyncClient
            if bool(self.files) is False:
                response = await client.post(url, json=self.json,
                                             params={'wait': True},
                                             timeout=self.timeout)
            else:
                self.files["payload_json"] = (None, json.dumps(self.json).encode('utf-8'))
                response = await client.post(url, files=self.files,
                                             timeout=self.timeout)
        return response

    async def execute(self, remove_embeds=False, remove_files=False):
        """
        executes the Webhook
        :param remove_embeds: if set to True, calls `self.remove_embeds()` to empty `self.embeds` after webhook is executed
        :param remove_files: if set to True, calls `self.remove_files()` to empty `self.files` after webhook is executed
        :return: Webhook response
        """
        webhook_urls = self.url if isinstance(self.url, list) else [self.url]
        urls_len = len(webhook_urls)
        responses = []
        for i, url in enumerate(webhook_urls):
            response = await self.api_post_request(url)
            if response.status_code in [200, 204]:
                logger.debug(
                    "[{index}/{length}] Webhook executed".format(
                        index=i + 1, length=urls_len
                    )
                )
            elif response.status_code == 429 and self.rate_limit_retry:
                while response.status_code == 429:
                    await self.handle_rate_limit(response)
                    response = await self.api_post_request(url)
                    if response.status_code in [200, 204]:
                        logger.debug(
                            "[{index}/{length}] Webhook executed".format(
                                index=i + 1, length=urls_len
                            )
                        )
                        break
            else:
                logger.error(
                    "[{index}/{length}] Webhook status code {status_code}: {content}".format(
                        index=i + 1,
                        length=urls_len,
                        status_code=response.status_code,
                        content=response.content.decode("utf-8"),
                    )
                )
            responses.append(response)
        if remove_embeds:
            self.remove_embeds()
        if remove_files:
            self.remove_files()
        return responses[0] if len(responses) == 1 else responses

    async def edit(self, sent_webhook):
        """
        edits the webhook passed as a response
        :param sent_webhook: webhook.execute() response
        :return: Another webhook response
        """
        sent_webhook = sent_webhook if isinstance(sent_webhook, list) else [sent_webhook]
        webhook_len = len(sent_webhook)
        responses = []
        async with self.http_client as client:  # type: httpx.AsyncClient
            for i, webhook in enumerate(sent_webhook):
                previous_sent_message_id = json.loads(webhook.content.decode('utf-8'))['id']
                url = webhook.url.split('?')[0] + '/messages/' + str(previous_sent_message_id)  # removes any query params
                if bool(self.files) is False:
                    patch_kwargs = {'json': self.json, 'params': {'wait': True}, 'timeout': self.timeout}
                else:
                    self.files["payload_json"] = (None, json.dumps(self.json))
                    patch_kwargs = {'files': self.files, 'timeout': self.timeout}
                response = await client.patch(url, **patch_kwargs)
                if response.status_code in [200, 204]:
                    logger.debug(
                        "[{index}/{length}] Webhook edited".format(
                            index=i + 1,
                            length=webhook_len,
                        )
                    )
                elif response.status_code == 429 and self.rate_limit_retry:
                    while response.status_code == 429:
                        await self.handle_rate_limit(response)
                        response = await client.patch(url, **patch_kwargs)
                        if response.status_code in [200, 204]:
                            logger.debug(
                                "[{index}/{length}] Webhook edited".format(
                                    index=i + 1,
                                    length=webhook_len,
                                )
                            )
                            break
                else:
                    logger.error(
                        "[{index}/{length}] Webhook status code {status_code}: {content}".format(
                            index=i + 1,
                            length=webhook_len,
                            status_code=response.status_code,
                            content=response.content.decode("utf-8"),
                        )
                    )
                responses.append(response)
        return responses[0] if len(responses) == 1 else responses

    async def delete(self, sent_webhook):
        """
        deletes the webhook passed as a response
        :param sent_webhook: webhook.execute() response
        :return: Response
        """
        sent_webhook = sent_webhook if isinstance(sent_webhook, list) else [sent_webhook]
        webhook_len = len(sent_webhook)
        responses = []
        async with self.http_client as client:  # type: httpx.AsyncClient
            for i, webhook in enumerate(sent_webhook):
                url = webhook.url.split('?')[0]  # removes any query params
                previous_sent_message_id = json.loads(webhook.content.decode('utf-8'))['id']
                response = await client.delete(url + '/messages/' + str(previous_sent_message_id), timeout=self.timeout)
                if response.status_code in [200, 204]:
                    logger.debug(
                        "[{index}/{length}] Webhook deleted".format(
                            index=i + 1,
                            length=webhook_len,
                        )
                    )
                else:
                    logger.error(
                        "[{index}/{length}] Webhook status code {status_code}: {content}".format(
                            index=i + 1,
                            length=webhook_len,
                            status_code=response.status_code,
                            content=response.content.decode("utf-8"),
                        )
                    )
                responses.append(response)
        return responses[0] if len(responses) == 1 else responses

    async def handle_rate_limit(self, response):
        """
        handles the rate limit
        :param response: Response
        :return: Response
        """
        errors = response.json()
        wh_sleep = (int(errors['retry_after']) / 1000) + 0.15
        await asyncio.sleep(wh_sleep)
        logger.error(
            "Webhook rate limited: sleeping for {wh_sleep} "
            "seconds...".format(
                wh_sleep=wh_sleep
            )
        )



from typing import Union


class ColorNotInRangeException(Exception):
    """
    A valid color must take an integer value between 0 and 16777216 inclusive

    This Exception will be raised when a colour is not in that range.
    """

    color: Union[str, int]

    def __init__(self, color: Union[str, int]) -> None:
        self.color = color
        super().__init__()

    def __str__(self) -> str:
        return repr(
            f"{self.color!r} is not in valid range of colors. The valid ranges "
            "of colors are 0 to 16777215 inclusive (INTEGERS) and 0 "
            "to FFFFFF inclusive (HEXADECIMAL)"
        )



import unittest

from discord_webhook.webhook import DiscordWebhook


class TestHermesApi(unittest.TestCase):
    """Tests to exercise webhooks"""

    def test_discordwebhook_constructor(self):
        webhook = DiscordWebhook("testurl")
        self.assertEqual(webhook.url, "testurl")

    def test_set_content(self):
        test_content = "Test data"
        webhook = DiscordWebhook("testurl")
        webhook.set_content(test_content)
        self.assertEqual(webhook.content, test_content)



import os

from setuptools import find_packages, setup

with open(os.path.join(os.path.dirname(__file__), "README.md")) as readme:
    README = readme.read()

# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

setup(
    name="discord-webhook",
    version="0.17.0",
    packages=find_packages(),
    include_package_data=True,
    license="MIT License",
    description="execute discord webhooks",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/lovvskillz/python-discord-webhook",
    install_requires=[
        "requests>=2.19.1",
    ],
    author="Vadim Zifra",
    author_email="vadim@minehub.de",
    extras_require={
        'async': [
            'httpx>=0.20.0'
        ]
    },
    classifiers=[
        "Environment :: Web Environment",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Internet :: WWW/HTTP :: Dynamic Content",
    ],
    entry_points={
        "console_scripts": [
            "discord_webhook=discord_webhook.__main__:main",
        ],
    },
)
